#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include "heap.h"
#include "traverse.h"

typedef struct File {
   char* path;
   time_t mtime;
} File;

int compare_oldest(void* a, void* b) {
   time_t time_a = ((File*) a)->mtime;
   time_t time_b = ((File*) b)->mtime;
   return time_a > time_b? 1: time_a < time_b? -1: 0;
}

int compare_newest(void* a, void* b) {
   time_t time_a = ((File*) a)->mtime;
   time_t time_b = ((File*) b)->mtime;
   return time_b > time_a? 1: time_b < time_a? -1: 0;
}

static bool newest_flag = true;

bool handler(void* handle, char* path, struct stat* statbuf) {
   Heap* heap = (Heap*) handle;

   if (!S_ISREG(statbuf->st_mode)) return true;

   bool filled = get_heap_size(heap) == get_heap_maxsize(heap);
   if (filled) {
      bool insert = false;
      time_t current_max = ((File*)get_max_from_heap(heap))->mtime;
      if (newest_flag) {
	 insert = statbuf->st_mtime > current_max;
      } else {
	 insert = statbuf->st_mtime < current_max;
      }
      if (!insert) return true; 
   }
   if (filled) {
      /* remove the newest file from the heap */
      File* file = extract_max_from_heap(heap);
      free(file->path); free(file);
   }
   File* file = malloc(sizeof(File));
   file->path = strdup(path); file->mtime = statbuf->st_mtime;
   insert_into_heap(heap, file);
   return true;
}

int main(int argc, char** argv) {
   char* cmdname = *argv;
   char usage[] = "Usage: %s [-c count] [-d] [-o] [dirpath]\n";
   int count = 1;
   bool date_flag = false;

   int option;
   while ((option = getopt(argc, argv, "c:do")) != -1) {
      switch (option) {
	 case 'c':
	    count = atoi(optarg);
	    if (count <= 0) {
	       fprintf(stderr, usage, cmdname); exit(1);
	    }
	    break;
	 case 'd':
	    date_flag = true;
	    break;
	 case 'o':
	    newest_flag = false;
	    break;
	 default:
	    fprintf(stderr, usage, cmdname); exit(1);
	    break;
      }
   }

   argv += optind; argc -= optind;
   if (argc > 1) {
      fprintf(stderr, usage, cmdname); exit(1);
   }
   char* path = ".";
   if (argc > 0) path = argv[0];

   Heap* heap;
   if (newest_flag) {
      heap = create_heap(count, compare_newest);
   } else {
      heap = create_heap(count, compare_oldest);
   }
   if (!traverse(path, heap, handler)) {
      perror(path); exit(1);
   }

   File* file;
   while ((file = extract_max_from_heap(heap))) {
      if (date_flag) {
	 char* timestr = ctime(&file->mtime);
	 char* cp = strchr(timestr, '\n');
	 if (cp) *cp = 0;
	 printf("%s %s\n", timestr, file->path);
      } else {
	 printf("%s\n", file->path);
      }
      free(file->path); free(file);
   }
}
